"use strict";
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    var desc = Object.getOwnPropertyDescriptor(m, k);
    if (!desc || ("get" in desc ? !m.__esModule : desc.writable || desc.configurable)) {
      desc = { enumerable: true, get: function() { return m[k]; } };
    }
    Object.defineProperty(o, k2, desc);
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);
    __setModuleDefault(result, mod);
    return result;
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.DynamodbEventStore = void 0;
const core_1 = require("@ddes/core");
const aws_sdk_1 = require("aws-sdk");
const orchestration = __importStar(require("./orchestration"));
/**
 * Interface for EventStore powered by DynamoDB
 */
class DynamodbEventStore extends core_1.EventStore {
    constructor(config) {
        super();
        const { tableName, client, initialCapacity = {
            tableRead: 2,
            tableWrite: 2,
            chronologicalRead: 2,
            chronologicalWrite: 2,
            instancesRead: 1,
            instancesWrite: 1,
        }, autoscaling, } = config;
        if (!tableName) {
            throw new Error(`'tableName' must be specified`);
        }
        this.client = client;
        this.tableName = tableName;
        this.initialCapacity = initialCapacity;
        this.autoscaling = autoscaling;
    }
    /**
     * Create DynamoDB table and auto scaling configuration
     */
    async setup() {
        await orchestration.createTable(this.tableSpecification, {
            dynamodb: this.client,
            ttl: true,
        });
        if (this.autoscaling) {
            await orchestration.setupAutoScaling(this.tableName, this.autoscaling);
        }
    }
    /** @inheritdoc */
    async teardown() {
        if (this.autoscaling) {
            await orchestration.removeAutoScaling(this.tableName);
        }
        await orchestration.deleteTable(this.tableName, {
            dynamodb: this.client,
        });
    }
    /**
     * @inheritdoc
     */
    async commit(commit) {
        try {
            await this.client
                .putItem({
                TableName: this.tableName,
                Item: this.marshallCommit(commit),
                ConditionExpression: 'attribute_not_exists(v)',
                ReturnValues: 'NONE',
            })
                .promise();
        }
        catch (error) {
            if (error.code === 'ConditionalCheckFailedException') {
                throw new core_1.VersionConflictError(commit);
            }
            throw error;
        }
        return commit;
    }
    /** @inheritdoc */
    async getAggregateHeadCommit(type, key) {
        for await (const commits of this.queryAggregateCommits(type, key, {
            descending: true,
            limit: 1,
        })) {
            for await (const commit of commits) {
                return commit;
            }
        }
    }
    /**
     * Retrieve commits from the store chronologically
     */
    async *chronologicalQuery(params) {
        var _a, _b, _c;
        const { group = 'default', min, descending, limit, exclusiveMin, exclusiveMax, timeDriftCompensation = 500, } = params;
        const { max = new Date(Date.now() + timeDriftCompensation) } = params;
        if (!min) {
            throw new Error('You must specify the "min" parameter');
        }
        const maxDate = max instanceof Date
            ? max
            : new Date(max.replace(/^(\d{4})(\d{2})(\d{2}).*/, '$1-$2-$3'));
        const maxCursor = max instanceof Date ? `${max.toISOString().replace(/[^0-9]/g, '')};` : max;
        const minDate = min instanceof Date
            ? min
            : new Date(min.replace(/^(\d{4})(\d{2})(\d{2}).*/, '$1-$2-$3'));
        const minCursor = min instanceof Date ? min.toISOString().replace(/[^0-9]/g, '') : min;
        let commitCount = 0;
        for (const partition of chronologicalPartitionIterator({
            start: minDate,
            end: maxDate,
            group,
            descending,
        })) {
            const queryParams = {
                IndexName: 'chronological',
                keyExpressions: ['p = :p', 'g BETWEEN :min AND :max'],
                filterExpressions: params.aggregateTypes
                    ? [
                        `a IN (${params.aggregateTypes.map((aggregateType, i) => `:a${i}`)})`,
                    ]
                    : undefined,
                queryVariables: {
                    ':p': partition.key,
                    ':min': exclusiveMin ? stringcrementor(minCursor) : minCursor,
                    ':max': exclusiveMax ? stringcrementor(maxCursor, -1) : maxCursor,
                    ...(_a = params.aggregateTypes) === null || _a === void 0 ? void 0 : _a.reduce((vars, type, i) => Object.assign(vars, {
                        [`:a${i}`]: type,
                    }), {}),
                },
                ScanIndexForward: !descending,
                ...(limit && { Limit: limit - commitCount }),
            };
            for await (const queryResult of this.request('query', queryParams)) {
                commitCount += queryResult.Items ? queryResult.Items.length : 0;
                if ((_b = queryResult.Items) === null || _b === void 0 ? void 0 : _b.length) {
                    yield (_c = queryResult.Items) === null || _c === void 0 ? void 0 : _c.map((i) => this.unmarshallCommit(i));
                }
                if (limit && commitCount >= limit) {
                    return;
                }
            }
        }
    }
    /**
     * Retrieve ordered commits for each aggregate instance of [[AggregateType]]
     */
    async *scanAggregateCommitsGroupedByKey(type) {
        const keyExpressions = ['a = :a'];
        const queryVariables = {
            ':a': type,
        };
        const filterExpressions = [];
        for await (const instanceQueryResult of this.request('query', {
            IndexName: 'instances',
            keyExpressions,
            queryVariables,
            filterExpressions,
        })) {
            if (instanceQueryResult.Items) {
                for (const rawItem of instanceQueryResult.Items) {
                    const { a, r } = aws_sdk_1.DynamoDB.Converter.unmarshall(rawItem);
                    for await (const commits of this.queryAggregateCommits(a, r)) {
                        yield commits;
                    }
                }
            }
        }
    }
    /**
     * Query the commits of an [[Aggregate]] instance
     */
    async *queryAggregateCommits(type, key, options = {}) {
        var _a;
        const { consistentRead = true, minVersion = 1, maxVersion = Number.MAX_SAFE_INTEGER, maxTime, descending, limit, } = options;
        if (!type || !key) {
            throw new Error('You need to specify "type" and "key"');
        }
        const keyExpressions = [
            's = :streamId',
            'v BETWEEN :minVersion and :maxVersion',
        ];
        const queryVariables = {
            ':streamId': [type, key].join(':'),
            ':minVersion': minVersion,
            ':maxVersion': maxVersion,
        };
        const filterExpressions = [];
        if (maxTime) {
            filterExpressions.push('t <= :maxTime');
            queryVariables[':maxTime'] = maxTime.valueOf();
        }
        const queryParams = {
            limit,
            ConsistentRead: consistentRead,
            ...(descending && { ScanIndexForward: false }),
        };
        const query = this.request('query', {
            keyExpressions,
            filterExpressions,
            queryVariables,
            ...queryParams,
        });
        for await (const result of query) {
            if ((_a = result.Items) === null || _a === void 0 ? void 0 : _a.length) {
                yield result.Items.map((item) => this.unmarshallCommit(item));
            }
        }
    }
    /**
     * Scan store commits
     */
    async *scan(params) {
        var _a;
        const { segment = 0, totalSegments = 1, ...rest } = params || {};
        const scan = this.request('scan', {
            TotalSegments: totalSegments,
            Segment: segment,
            ...rest,
        });
        for await (const result of scan) {
            if ((_a = result.Items) === null || _a === void 0 ? void 0 : _a.length) {
                yield result.Items.map((item) => this.unmarshallCommit(item));
            }
        }
    }
    streamCommits(params, yieldEmpty) {
        // TODO
        return undefined;
    }
    streamAggregateInstanceCommits(aggregateType, key, minVersion = 1, yieldEmpty) {
        // TODO
        return undefined;
    }
    //
    // PROTECTED
    //
    chronologicalKey(data) {
        return [
            new Date(data.timestamp).toISOString().replace(/[^0-9]/g, ''),
            data.aggregateType,
            data.aggregateKey,
            data.aggregateVersion,
        ].join(':');
    }
    marshallCommit(commit) {
        const { aggregateType, aggregateKey, events, timestamp, aggregateVersion, expiresAt, chronologicalPartition = 'default', chronologicalKey, } = commit;
        return aws_sdk_1.DynamoDB.Converter.marshall({
            s: [aggregateType, aggregateKey].join(':'),
            v: aggregateVersion,
            g: chronologicalKey,
            a: aggregateType,
            r: aggregateVersion === 1 ? aggregateKey : undefined,
            t: new Date(timestamp).valueOf(),
            e: JSON.stringify(events),
            x: expiresAt,
            p: `${new Date(timestamp)
                .toISOString()
                .split('T')[0]
                .replace(/\-/g, '')}${chronologicalPartition}`,
        });
    }
    unmarshallCommit(marshalledCommit) {
        const unmarshalled = aws_sdk_1.DynamoDB.Converter.unmarshall(marshalledCommit);
        const [, aggregateType, aggregateKey] = unmarshalled.s.match(/^([^:]*):(.*)$/);
        const commit = {
            aggregateType,
            aggregateKey,
            aggregateVersion: unmarshalled.v,
            expiresAt: unmarshalled.x,
            timestamp: unmarshalled.t,
            events: JSON.parse(unmarshalled.e),
            chronologicalPartition: unmarshalled.p.substr(8, unmarshalled.p.length - 8),
            chronologicalKey: unmarshalled.g,
        };
        return commit;
    }
    get tableSpecification() {
        return {
            TableName: this.tableName,
            AttributeDefinitions: [
                { AttributeName: 's', AttributeType: 'S' },
                { AttributeName: 'v', AttributeType: 'N' },
                { AttributeName: 'g', AttributeType: 'S' },
                { AttributeName: 'p', AttributeType: 'S' },
                { AttributeName: 'a', AttributeType: 'S' },
                { AttributeName: 'r', AttributeType: 'S' },
            ],
            KeySchema: [
                { AttributeName: 's', KeyType: 'HASH' },
                { AttributeName: 'v', KeyType: 'RANGE' },
            ],
            ProvisionedThroughput: {
                ReadCapacityUnits: this.initialCapacity.tableRead,
                WriteCapacityUnits: this.initialCapacity.tableWrite,
            },
            GlobalSecondaryIndexes: [
                {
                    IndexName: 'chronological',
                    KeySchema: [
                        { AttributeName: 'p', KeyType: 'HASH' },
                        { AttributeName: 'g', KeyType: 'RANGE' },
                    ],
                    Projection: {
                        ProjectionType: 'INCLUDE',
                        NonKeyAttributes: ['t', 'e', 'x', 'a'],
                    },
                    ProvisionedThroughput: {
                        ReadCapacityUnits: this.initialCapacity.chronologicalRead,
                        WriteCapacityUnits: this.initialCapacity.chronologicalWrite,
                    },
                },
                {
                    IndexName: 'instances',
                    KeySchema: [
                        { AttributeName: 'a', KeyType: 'HASH' },
                        { AttributeName: 'r', KeyType: 'RANGE' },
                    ],
                    Projection: {
                        ProjectionType: 'KEYS_ONLY',
                    },
                    ProvisionedThroughput: {
                        ReadCapacityUnits: this.initialCapacity.instancesRead,
                        WriteCapacityUnits: this.initialCapacity.instancesWrite,
                    },
                },
            ],
        };
    }
    async *request(type, params = {}) {
        const { keyExpressions = [], filterExpressions = [], queryVariables = {}, limit, ...additionalQueryParams } = params;
        const queryParams = {
            TableName: this.tableName,
            ReturnConsumedCapacity: 'TOTAL',
            ...(keyExpressions.length && {
                KeyConditionExpression: keyExpressions.join(' AND '),
            }),
            ...(filterExpressions.length && {
                FilterExpression: filterExpressions.join(' AND '),
            }),
            ...(Object.keys(queryVariables).length && {
                ExpressionAttributeValues: aws_sdk_1.DynamoDB.Converter.marshall(queryVariables),
            }),
            ...(limit && { Limit: limit }),
            ...additionalQueryParams,
        };
        let lastEvaluatedKey;
        do {
            let request;
            if (type === 'scan') {
                request = this.client.scan({
                    ...queryParams,
                    ...(lastEvaluatedKey ? { ExclusiveStartKey: lastEvaluatedKey } : {}),
                });
            }
            else if (type === 'query') {
                request = this.client.query({
                    ...queryParams,
                    ...(lastEvaluatedKey ? { ExclusiveStartKey: lastEvaluatedKey } : {}),
                });
            }
            else {
                throw new Error('Invalid request type');
            }
            const result = await request.promise();
            yield result;
            lastEvaluatedKey = result.LastEvaluatedKey;
        } while (lastEvaluatedKey);
    }
}
exports.DynamodbEventStore = DynamodbEventStore;
function stringcrementor(str, step = 1) {
    const buffer = Buffer.from(str);
    if (step === 1) {
        if (buffer[buffer.length - 1] === 255) {
            return Buffer.concat([buffer, Buffer.from(' ')]).toString();
        }
        buffer[buffer.length - 1]++;
        return buffer.toString();
    }
    if (step === -1) {
        if (buffer[buffer.length - 1] <= 32) {
            return buffer.subarray(0, buffer.length - 1).toString();
        }
        buffer[buffer.length - 1]--;
        return buffer.toString();
    }
    return '';
}
function* chronologicalPartitionIterator(params) {
    const { start, end = new Date(), group = 'default', descending } = params;
    let partitionCursor = new Date(descending ? end : start);
    partitionCursor.setUTCHours(0, 0, 0, 0);
    while (descending ? partitionCursor >= start : partitionCursor <= end) {
        const endsAt = new Date(partitionCursor);
        endsAt.setUTCHours(23, 59, 59, 999);
        yield {
            key: `${partitionCursor
                .toISOString()
                .split('T')[0]
                .replace(/\-/g, '')}${group}`,
            startsAt: partitionCursor,
            endsAt,
        };
        partitionCursor = new Date(partitionCursor.valueOf() + 24 * 60 * 60 * 1000 * (descending ? -1 : 1));
    }
}
//# sourceMappingURL=DynamodbEventStore.js.map